#!/usr/bin/perl

use Geo::ShapeFile::Point comp_includes_m => 0, comp_includes_z => 0;
use Geo::ShapeFile;
use List::Util qw(sum);
use Math::GSL::Statistics qw /:all/;
use Geo::ShapeFile::Shape qw /:all/;
use Math::Polygon;

use strict;

my %mesh_block;
my @data;

sub LoadMeshData
{
	die("Unable to open $_[0]!\n") unless open (IN, $_[0]);

	while (<IN>)
	{
		next if $_ =~ /#.*/;
		my @item = split /[,\n]/;
		next unless $item[0] =~ /\s*(MB)\s*(\d+)/;

		my $id = $1.$2;
		if (defined $mesh_block{$id})
		{
			$mesh_block{$id}{'population'} = $item[3];
			push @data, $item[3];
		}
	}
	
	close IN;
}

die ("Wrong parameters!") unless @ARGV == 3;

my ($mesh_file, $shape_file, $output_file) = @ARGV;

my $shapefile = new Geo::ShapeFile($shape_file);

print "Shape type: ", $shapefile->shape_type_text(), "\n";

for(1 .. $shapefile->shapes()) 
{
	my $shape = $shapefile->get_shp_record($_);
# see Geo::ShapeFile::Shape docs for what to do with $shape
	die "Non-single part!" unless $shape->num_parts() == 1;

	die "Non-polygon!\n" unless $shapefile->type($shape->shape_type()) eq "Polygon";

	my %db = $shapefile->get_dbf_record($_);

	foreach my $key (keys %db)
	{
		if ($key =~ /MB.*/)
		{
			my $id = "MB".$db{$key};
			if (defined $mesh_block{$id})
			{
				print "Existing mesh block: $id renamed to "; 
				$mesh_block{$id}{'count'}++;
				$id.= "-".$mesh_block{$id}{'count'}; 
				print "$id\n";
			}
			else
			{
				$mesh_block{$id}{'count'} = 1;
			}
			my $first_point = @{$shape->points()}[0];
			my $last_point;
			foreach my $point ($shape->points())
			{
				$last_point = $point;
				push @{$mesh_block{$id}{'points'}}, [$point->X(), $point->Y()];
			}

                        $mesh_block{$id}{'shape'} = $shape;
		}
	}
}
print "Number of extracted polygons: ", $shapefile->shapes(), "\n";

print "Loading mesh data...";
LoadMeshData($mesh_file);
print "done\n";

print "Number of meshblocks: ", scalar(@data), "\n";
print "Number of people: ", sum(@data), "\n";
print "Mean meshblock population: ", gsl_stats_mean(\@data, 1, scalar(@data)), "\n";
print "Stddev of meshblock population: ", gsl_stats_sd(\@data, 1, scalar(@data)), "\n";
print "Min meshblock population: ", gsl_stats_min(\@data, 1, scalar(@data)), "\n";
print "Max meshblock population: ", gsl_stats_max(\@data, 1, scalar(@data)), "\n";

# Now we have the geometrics and the population of each mesh block

die ("Unable to open output file!") unless open(OUT, ">$output_file");
# This needs for the original simulation!!!
#print OUT $shapefile->x_max - $shapefile->x_min,"\t",$shapefile->y_max - $shapefile->y_min,"\t",$shapefile->x_max - $shapefile->x_min,"\t",
#      $shapefile->y_max - $shapefile->y_min,"\n";

my $block_num = scalar(keys %mesh_block);
my $i = 0;
$|=1;
foreach my $block (keys %mesh_block)
{
	$i++;
	print "\rGenerating random population in block $block ($i/$block_num)     ";
	my $poly = Math::Polygon->new( @{$mesh_block{$block}{'points'}} );

	my $shape = $mesh_block{$block}{'shape'};
#	print "$x_min, $x_max, $y_min, $y_max\n";
#	print "ID: $block\n";

	for (1..$mesh_block{$block}{'population'})
	{
		my ($x, $y);
		do  
		{
			$x = int(rand($shape->x_max - $shape->x_min)) + $shape->x_min;
			$y = int(rand($shape->y_max - $shape->y_min)) + $shape->y_min;
		}
		while (!$poly->contains([$x,$y]));
		print OUT "$x\t$y\n";
	}
}
close OUT;
print "\n";






